'''Crash database interface for crashdb.opensuse.org

TODO:
- detect multiple identical uploads on the server side
- prevent that anything else than an apport-report is uploaded to the server
- add valid SSL certificate for crashdb.o.o
'''

from MultipartPostHandler import MultipartPostHandler
from HTTPSValidateCertificateConnection import *
from HTTPProgressConnection import HTTPProgressHandler
import tempfile, urllib2, cookielib

import apport.crashdb
import apport

class CrashDatabase(apport.crashdb.CrashDatabase):

    def __init__(self, auth_file, bugpattern_baseurl, options):
        '''Initialize crash database connection. 
        
        You need to specify an implementation specific file with the
        authentication credentials for retracing access for download() and
        update(). For upload() and get_comment_url() you can use None.
        
        options is a dictionary with additional settings from crashdb.conf; see
        get_crashdb() for details'''

        apport.crashdb.CrashDatabase.__init__(self, auth_file,
                                              bugpattern_baseurl, options)

        self.report_baseurl = options.get('report_baseurl')
        self.comment_baseurl = options.get('comment_baseurl')
        self.comment_rawurl = options.get('comment_rawurl')

        if not self.report_baseurl:
            raise RuntimeError, "No report_baseurl defined"

        HTTPSValidateCertificateConnection.set_cert_location(options.get('cert_location'))

    def upload(self, report, progress_callback = None):
        '''Upload given problem report return a handle for it. 
        
        This should happen noninteractively.'''

        cookies = cookielib.CookieJar()
        progress_handler = HTTPProgressHandler(progress_callback)
        opener = urllib2.build_opener(urllib2.HTTPCookieProcessor(cookies),
                                      HTTPSValidateCertificateHandler,
                                      progress_handler,
                                      MultipartPostHandler)
        
        tmp = tempfile.TemporaryFile()
        report.write(tmp)
        tmp.seek(0)

        result = opener.open(self.report_baseurl,
                             { 'FORM_SUBMIT': '1', 'report': tmp })

        if result.info().has_key('X-Apport-Error'):
            raise urllib2.HTTPError(result.geturl(), 400,
                                    result.info().get('X-Apport-Error'),
                                    result.headers, result.fp)

        uuid = result.info().get('X-Apport-Report-UUID')
        assert uuid
        return uuid

    def get_comment_url(self, report, handle):
        '''Return an URL that should be opened after report has been uploaded
        and upload() returned handle.

        Should return None if no URL should be opened (anonymous filing without
        user comments); in that case this function should do whichever
        interactive steps it wants to perform.'''

        if self.comment_baseurl:
            return self.comment_baseurl % ( handle )
        else:
            return None

    def download(self, id):
        '''Download the problem report from given ID and return a Report.'''

        url = self.comment_rawurl % ( id )
        cookies = cookielib.CookieJar()
        opener = urllib2.build_opener(urllib2.HTTPCookieProcessor(cookies),
                                      HTTPSValidateCertificateHandler)
        try:
            result = opener.open(url)
        except urllib2.HTTPError:
            return None
        except urllib2.URLError:
            return None

        encoding = result.headers.get('Content-Encoding')
        tmp = tempfile.TemporaryFile()
        while 1:
            lines = result.readlines(8*1024)
            if not lines: break
            for line in lines:
                if encoding != None:
                    tmp.write(line.decode(encoding))
                else:
                    tmp.write(line)
        tmp.seek(0)
        report = apport.Report()
        report.load(tmp)
        return report

    def update(self, id, report, comment):
        '''Update the given report ID with the retraced results from the report
        (Stacktrace, ThreadStacktrace, StacktraceTop; also Disassembly if
        desired) and an optional comment.'''

        raise NotImplementedError, 'this method must be implemented by a concrete subclass'

    def get_distro_release(self, id):
        '''Get 'DistroRelease: <release>' from the given report ID and return
        it.'''

        raise NotImplementedError, 'this method must be implemented by a concrete subclass'

    def get_unretraced(self):
        '''Return an ID set of all crashes which have not been retraced yet and
        which happened on the current host architecture.'''

        raise NotImplementedError, 'this method must be implemented by a concrete subclass'

    def get_dup_unchecked(self):
        '''Return an ID set of all crashes which have not been checked for
        being a duplicate.

        This is mainly useful for crashes of scripting languages such as
        Python, since they do not need to be retraced. It should not return
        bugs that are covered by get_unretraced().'''

        raise NotImplementedError, 'this method must be implemented by a concrete subclass'

    def get_unfixed(self):
        '''Return an ID set of all crashes which are not yet fixed.

        The list must not contain bugs which were rejected or duplicate.
        
        This function should make sure that the returned list is correct. If
        there are any errors with connecting to the crash database, it should
        raise an exception (preferably IOError).'''

        raise NotImplementedError, 'this method must be implemented by a concrete subclass'

    def get_fixed_version(self, id):
        '''Return the package version that fixes a given crash.

        Return None if the crash is not yet fixed, or an empty string if the
        crash is fixed, but it cannot be determined by which version. Return
        'invalid' if the crash report got invalidated, such as closed a
        duplicate or rejected.

        This function should make sure that the returned result is correct. If
        there are any errors with connecting to the crash database, it should
        raise an exception (preferably IOError).'''

        raise NotImplementedError, 'this method must be implemented by a concrete subclass'

    def close_duplicate(self, id, master):
        '''Mark a crash id as duplicate of given master ID.'''

        raise NotImplementedError, 'this method must be implemented by a concrete subclass'

    def mark_regression(self, id, master):
        '''Mark a crash id as reintroducing an earlier crash which is
        already marked as fixed (having ID 'master').'''
        
        raise NotImplementedError, 'this method must be implemented by a concrete subclass'

    def mark_retraced(self, id):
        '''Mark crash id as retraced.'''

        raise NotImplementedError, 'this method must be implemented by a concrete subclass'

    def mark_retrace_failed(self, id):
        '''Mark crash id as 'failed to retrace'.
        
        This can be a no-op if you are not interested in this, of course.'''

        raise NotImplementedError, 'this method must be implemented by a concrete subclass'

    def _mark_dup_checked(self, id, report):
        '''Mark crash id as checked for being a duplicate
        
        This is an internal method that should not be called from outside.'''

        raise NotImplementedError, 'this method must be implemented by a concrete subclass'

###################################################################

if __name__=="__main__":
    import unittest
    from cStringIO import StringIO

    class _OpensuseCrashDBTest(unittest.TestCase):
        def setUp(self):
            if True:
                self.crashes = CrashDatabase(None, None,
                                             { 'cert_location':
                                                   '/etc/apport/crashdb.crt',
                                               'report_baseurl':
                                                   'https://crashdb.opensuse.org/log/new',
                                               'comment_baseurl':
                                                   'https://crashdb.opensuse.org/log/show/%s',
                                               'comment_rawurl':
                                                   'https://crashdb.opensuse.org/log/raw/%s',
                                               })
            else:
                # For testing with a local server installation
                self.crashes = CrashDatabase(None, None,
                                             { 'cert_location':
                                                   'crashdb.crt',
                                               'report_baseurl':
                                                   'http://localhost:3000/log/new',
                                               'comment_baseurl':
                                                   'http://localhost:3000/log/show/%s',
                                               'comment_rawurl':
                                                   'http://localhost:3000/log/raw/%s',
                                               })

        def test_download(self):
            report = self.crashes.download('00000000000000000000000000000000')
            self.assert_(report != None, "Couldn't download test report")
            self.assertEqual(report['Date'], 'Mon Oct 20 09:55:27 2008')

        def _md5sum(self, report):
            try:
                import hashlib
                m = hashlib.md5()
            except ImportError:
                # for Python << 2.5
                import md5
                m = md5.new()

            io = StringIO()
            report.write(io)
            io.seek(0)

            while True:
                d = io.read(8096)
                if not d:
                    break
                m.update(d)

            return m.hexdigest()

        def test_upload_bugreport(self):

            report = apport.Report()

            report['ProblemType'] = 'Bug'
            report['Architecture'] = 'i386'
            report['Date'] = 'Thu Jan  1 01:00:00 CET 1970'
            report['Uname'] = 'Linux 0.1-1-default i386'
            report['ExecutablePath'] = '/bin/cat'
            report['ProcEnviron'] = 'PATH=/bin'
            report['ProcCmdline'] = '/bin/cat'
            report['ProcStatus'] = '''Name:   cat
State:  R (running)'''
            report['ProcMaps'] = '''00400000-0040a000 r-xp 00000000 fd:01 524521                             /bin/cat'''
            report['Package'] = 'coreutils'
            report['Dependencies'] = 'rpmlib(VersionedDependencies) <= 3.0.3-1'

            report['Ignore'] = 'True'

            try:
                uuid = self.crashes.upload(report)
            except urllib2.HTTPError, e:
                print "There was an http error: %s" % e
                print e.info()

            self.assertEqual(uuid, self._md5sum(report))

        def test_upload_crashreport(self):

            report = apport.Report()
            try:
                report.load(open('/suse/jblunck/Work/apport/reports/_usr_lib_gvfs_gvfsd-trash.16484.crash.processed'))
            except IOError:
                self.fail('Couldn\'t find test report')
            report['Ignore'] = 'True'

            try:
                uuid = self.crashes.upload(report)
            except urllib2.HTTPError, e:
                print "There was an http error: %s" % e
                print e.info()

            self.assertEqual(uuid, self._md5sum(report))

        def test_upload_emptyreport(self):

            report = apport.Report()

            try:
                uuid = self.crashes.upload(report)
            except urllib2.HTTPError, e:
                self.assert_(e.info().has_key('X-Apport-Error'))
                self.assertEqual(e.code, 400)
                return
            else:
                self.fail('Expected a HTTPError')

        def test_upload_emptyignorereport(self):

            report = apport.Report()
            report['Ignore'] = 'True'

            try:
                uuid = self.crashes.upload(report)
            except urllib2.HTTPError, e:
                self.assert_(e.info().has_key('X-Apport-Error'))
                self.assertEqual(e.code, 400)
                return
            else:
                self.fail('Expected a HTTPError')

        def test_upload_invalidreport(self):

            report = apport.Report()
            report['Foo'] = 'Bar'
            report['Ignore'] = 'True'

            try:
                uuid = self.crashes.upload(report)
            except urllib2.HTTPError, e:
                self.assert_(e.info().has_key('X-Apport-Error'))
                self.assertEqual(e.code, 400)
                return
            else:
                self.fail('Expected a HTTPError')

    unittest.main()
